@@ -1,4 +1,5 @@
 /*
+ * Copyright 2012-2022 Great Scott Gadgets <info@greatscottgadgets.com>
  * Copyright 2012 Jared Boone <jared@sharebrained.com>
  *
  * This file is part of HackRF.
@@ -24,9 +25,17 @@
 
 #include <stdint.h>
 
-volatile uint32_t* peripheral_bitband_address(volatile void* const address, const uint_fast8_t bit_number);
-void peripheral_bitband_set(volatile void* const peripheral_address, const uint_fast8_t bit_number);
-void peripheral_bitband_clear(volatile void* const peripheral_address, const uint_fast8_t bit_number);
-uint32_t peripheral_bitband_get(volatile void* const peripheral_address, const uint_fast8_t bit_number);
+volatile uint32_t* peripheral_bitband_address(
+	volatile void* const address,
+	const uint_fast8_t bit_number);
+void peripheral_bitband_set(
+	volatile void* const peripheral_address,
+	const uint_fast8_t bit_number);
+void peripheral_bitband_clear(
+	volatile void* const peripheral_address,
+	const uint_fast8_t bit_number);
+uint32_t peripheral_bitband_get(
+	volatile void* const peripheral_address,
+	const uint_fast8_t bit_number);
 
-#endif//__BITBAND_H__
+#endif //__BITBAND_H__
