@@ -1,4 +1,5 @@
 /*
+ * Copyright 2012-2022 Great Scott Gadgets <info@greatscottgadgets.com>
  * Copyright 2012 Jared Boone
  * Copyright 2013 Benjamin Vernoux
  *
@@ -26,33 +27,43 @@
 #include <usb_type.h>
 #include <usb_request.h>
 
-usb_request_status_t usb_vendor_request_write_max2837(
+usb_request_status_t usb_vendor_request_write_max283x(
 	usb_endpoint_t* const endpoint,
-	const usb_transfer_stage_t stage
-);
-usb_request_status_t usb_vendor_request_read_max2837(
+	const usb_transfer_stage_t stage);
+usb_request_status_t usb_vendor_request_read_max283x(
 	usb_endpoint_t* const endpoint,
-	const usb_transfer_stage_t stage
-);
+	const usb_transfer_stage_t stage);
 usb_request_status_t usb_vendor_request_write_si5351c(
 	usb_endpoint_t* const endpoint,
-	const usb_transfer_stage_t stage
-);
+	const usb_transfer_stage_t stage);
 usb_request_status_t usb_vendor_request_read_si5351c(
 	usb_endpoint_t* const endpoint,
-	const usb_transfer_stage_t stage
-);
+	const usb_transfer_stage_t stage);
 usb_request_status_t usb_vendor_request_write_rffc5071(
 	usb_endpoint_t* const endpoint,
-	const usb_transfer_stage_t stage
-);
+	const usb_transfer_stage_t stage);
 usb_request_status_t usb_vendor_request_read_rffc5071(
 	usb_endpoint_t* const endpoint,
-	const usb_transfer_stage_t stage
-);
+	const usb_transfer_stage_t stage);
 usb_request_status_t usb_vendor_request_set_clkout_enable(
 	usb_endpoint_t* const endpoint,
-	const usb_transfer_stage_t stage
-);
+	const usb_transfer_stage_t stage);
+usb_request_status_t usb_vendor_request_get_clkin_status(
+	usb_endpoint_t* const endpoint,
+	const usb_transfer_stage_t stage);
+usb_request_status_t usb_vendor_request_set_leds(
+	usb_endpoint_t* const endpoint,
+	const usb_transfer_stage_t stage);
+usb_request_status_t usb_vendor_request_user_config_set_bias_t_opts(
+	usb_endpoint_t* const endpoint,
+	const usb_transfer_stage_t stage);
+#ifdef PRALINE
+usb_request_status_t usb_vendor_request_write_fpga_reg(
+	usb_endpoint_t* const endpoint,
+	const usb_transfer_stage_t stage);
+usb_request_status_t usb_vendor_request_read_fpga_reg(
+	usb_endpoint_t* const endpoint,
+	const usb_transfer_stage_t stage);
+#endif
 
 #endif /* end of include guard: __USB_API_REGISTER_H__ */
